"""Main application class of eko."""
import logging
import os
from typing import Union

import numpy as np

from .. import interpolation
from ..couplings import Couplings, couplings_mod_ev
from ..evolution_operator.grid import OperatorGrid
from ..io import EKO, Operator, runcards
from ..io.types import RawCard, ScaleVariationsMethod
from ..thresholds import ThresholdsAtlas
from . import commons

logger = logging.getLogger(__name__)


class Runner:
    """Represents a single input configuration.

    For details about the configuration, see :doc:`here </code/IO>`

    Attributes
    ----------
    setup : dict
        input configurations

    """

    banner = commons.BANNER

    def __init__(
        self,
        theory_card: Union[RawCard, runcards.TheoryCard],
        operators_card: Union[RawCard, runcards.OperatorCard],
        path: os.PathLike,
    ):
        """Initialize runner.

        Parameters
        ----------
        theory_card :
            theory parameters and options
        operators_card :
            operator specific options
        path :
            path where to store the computed operator

        """
        new_theory, new_operator = runcards.update(theory_card, operators_card)

        # Store inputs
        self.path = path
        self._theory = new_theory

        # setup basis grid
        bfd = interpolation.InterpolatorDispatcher(
            xgrid=new_operator.xgrid,
            polynomial_degree=new_operator.configs.interpolation_polynomial_degree,
        )

        # setup the Threshold path, compute masses if necessary
        masses = runcards.masses(new_theory, new_operator.configs.evolution_method)

        # call explicitly iter to explain the static analyzer that is an
        # iterable
        thresholds_ratios = np.power(list(iter(new_theory.heavy.matching_ratios)), 2.0)
        tc = ThresholdsAtlas(
            masses=masses,
            q2_ref=new_operator.mu20,
            nf_ref=new_theory.heavy.num_flavs_init,
            thresholds_ratios=thresholds_ratios,
            max_nf=new_theory.heavy.num_flavs_max_pdf,
        )

        # strong coupling
        sc = Couplings(
            couplings=new_theory.couplings,
            order=new_theory.order,
            method=couplings_mod_ev(new_operator.configs.evolution_method),
            masses=masses,
            hqm_scheme=new_theory.heavy.masses_scheme,
            thresholds_ratios=thresholds_ratios
            * (
                new_theory.xif**2
                if new_operator.configs.scvar_method
                == ScaleVariationsMethod.EXPONENTIATED
                else 1.0
            ),
        )
        # setup operator grid
        self.op_grid = OperatorGrid(
            mu2grid=new_operator.mu2grid,
            order=new_theory.order,
            masses=masses,
            mass_scheme=new_theory.heavy.masses_scheme.value,
            intrinsic_flavors=new_theory.heavy.intrinsic_flavors,
            xif=new_theory.xif,
            configs=new_operator.configs,
            debug=new_operator.debug,
            thresholds_config=tc,
            couplings=sc,
            interpol_dispatcher=bfd,
            n3lo_ad_variation=new_theory.n3lo_ad_variation,
        )

        with EKO.create(path) as builder:
            builder.load_cards(  # pylint: disable=E1101
                new_theory, new_operator
            ).build()

    def compute(self):
        """Run evolution and generate output operator.

        Two steps are applied sequentially:

        1. evolution is performed, computing the evolution operator in an
           internal flavor and x-space basis
        2. bases manipulations specified in the runcard are applied

        """
        with EKO.edit(self.path) as eko:
            # add all operators
            for final_scale, op in self.op_grid.compute().items():
                eko[float(final_scale)] = Operator.from_dict(op)



"""IO generic exceptions."""
import os


class OutputError(Exception):
    """Generic Output Error."""


class OutputExistsError(FileExistsError, OutputError):
    """Output file already exists."""


class OutputNotTar(ValueError, OutputError):
    """Specified file is not a .tar archive."""


class OperatorLoadingError(ValueError, OutputError):
    """Issue encountered while loading an operator."""


class OperatorLocationError(ValueError, OutputError):
    """Path supposed to store an operator in wrong location."""

    def __init__(self, path: os.PathLike):
        self.path = path
        super().__init__(f"Path '{path}' not in operators folder")


class ReadOnlyOperator(RuntimeError, OutputError):
    """It is not possible to write on a read-only operator.

    In particular, the behavior would be deceitful, since writing is possible
    in-memory and even on the temporary folder.
    But eventually, no writing will happen on a persistent archive, so any
    modification is lost after exiting the program.

    """


class ClosedOperator(RuntimeError, OutputError):
    """It is not possible to write on nor to read from a closed operator.

    This is milder issue than :class:`ReadOnlyOperator`, since in this case not
    even the writing on the temporary folder would be possible.

    Instead, it will look like you can access some properties, but the operator
    is actually closed, so it should not be used any longer in general.
    However, for extremely simple properties, like those available in memory
    from :class:`eko.io.struct.Metadata` or :class:`eko.io.struct.AccessConfigs`, there
    is no need to raise on read, since those properties are actually available,
    but they should always raise on writing, since there is no persistence for
    the content written, and it can be deceitful.

    Still, the level of protection will be mild, since a thoruough protection
    would clutter a lot the code, requiring a lot of maintenance.
    "We are adult here".

    """



r"""The polarized, space-like |OME|."""

import numba as nb


@nb.njit(cache=True)
def A_non_singlet(_matching_order, _n, _nf, _L):
    """Compute the non-singlet |OME|."""
    raise NotImplementedError("Polarised, space-like is not yet implemented")


@nb.njit(cache=True)
def A_singlet(_matching_order, _n, _nf, _L, _is_msbar):
    """Compute the singlet |OME|."""
    raise NotImplementedError("Polarised, space-like is not yet implemented")



"""Operator components."""
from abc import ABC
from dataclasses import dataclass

import numpy.typing as npt

from ..io.dictlike import DictLike
from ..io.types import SquaredScale


@dataclass
class Part(DictLike, ABC):
    """An atomic operator ingredient.

    The operator is always in flavor basis, and on the x-grid specified for
    computation, i.e. it is a rank-4 tensor of shape::

        (flavor basis) x (x-grid) x (flavor basis) x (x-grid)

    """

    operator: npt.NDArray


@dataclass
class Evolution(Part):
    """Evolution in a fixed number of flavors."""

    final: bool
    mu20: SquaredScale
    mu2: SquaredScale


@dataclass
class Matching(Part):
    """Matching conditions between two different flavor schemes, at a given scale."""

    mu2: SquaredScale



r"""Holds the classes that define the |FNS|."""
import logging
from dataclasses import astuple, dataclass
from typing import List, Optional

import numpy as np

logger = logging.getLogger(__name__)


@dataclass
class PathSegment:
    """Oriented path in the threshold landscape."""

    q2_from: float
    """Starting point."""
    q2_to: float
    """Final point."""
    nf: int
    """Number of active flavors."""

    @property
    def is_downward_q2(self) -> bool:
        """Return True if ``q2_from`` is bigger than ``q2_to``."""
        return self.q2_from > self.q2_to

    @property
    def tuple(self):
        """Deprecated: use directly `dataclasses.astuple`."""
        return astuple(self)

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        return f"PathSegment({self.q2_from} -> {self.q2_to}, nf={self.nf})"


class ThresholdsAtlas:
    r"""Holds information about the matching scales.

    These scales are the :math:`Q^2` has to pass in order to get there from a
    given :math:`Q^2_{ref}`.

    """

    def __init__(
        self,
        masses: List[float],
        q2_ref: Optional[float] = None,
        nf_ref: Optional[int] = None,
        thresholds_ratios: Optional[List[float]] = None,
        max_nf: Optional[int] = None,
    ):
        """Create basic atlas.

        Parameters
        ----------
        masses :
            list of quark masses squared
        q2_ref :
            reference scale
        nf_ref :
            number of active flavors at the reference scale
        thresholds_ratios :
            list of ratios between matching scales and masses squared
        max_nf :
            maximum number of active flavors, if `None` no maximum is set

        """
        sorted_masses = sorted(masses)
        if not np.allclose(masses, sorted_masses):
            raise ValueError("masses need to be sorted")
        # combine them
        thresholds = self.build_area_walls(sorted_masses, thresholds_ratios, max_nf)
        self.area_walls = [0] + thresholds + [np.inf]

        # check nf_ref
        if nf_ref is not None:
            if q2_ref is None:
                raise ValueError(
                    "Without a reference Q2 value a reference number of flavors "
                    "does not make sense!"
                )
            # else self.q2_ref is not None
            nf_init = 2 + len(list(filter(lambda x: np.isclose(0, x), self.area_walls)))
            if nf_ref < nf_init:
                raise ValueError(
                    f"The reference number of flavors is set to {nf_ref}, "
                    f"but the atlas starts at {nf_init}"
                )
            nf_final = 2 + len(list(filter(lambda x: x < np.inf, self.area_walls)))
            if nf_ref > nf_final:
                raise ValueError(
                    f"The reference number of flavors is set to {nf_ref}, "
                    f"but the atlas stops at {nf_final}"
                )

        # Init values
        self.q2_ref = q2_ref
        self.nf_ref = nf_ref
        self.thresholds_ratios = thresholds_ratios
        logger.info(str(self))

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        walls = " - ".join([f"{w:.2e}" for w in self.area_walls])
        return f"ThresholdsAtlas [{walls}], ref={self.q2_ref} @ {self.nf_ref}"

    @classmethod
    def ffns(cls, nf: int, q2_ref: Optional[float] = None):
        """Create a |FFNS| setup.

        The function creates simply sufficient thresholds at ``0`` (in the
        beginning), since the number of flavors is determined by counting
        from below.

        Parameters
        ----------
        nf :
            number of light flavors
        q2_ref :
            reference scale

        """
        return cls([0] * (nf - 3) + [np.inf] * (6 - nf), q2_ref)

    @staticmethod
    def build_area_walls(
        masses: List[float],
        thresholds_ratios: Optional[List[float]] = None,
        max_nf: Optional[int] = None,
    ):
        r"""Create the object from the informations on the run card.

        The thresholds are computed by :math:`(m_q \cdot k_q^{Thr})`.

        Parameters
        ----------
        masses :
            heavy quark masses squared
        thresholds_ratios :
            list of ratios between matching scales and masses squared
        max_nf :
            maximum number of flavors

        Returns
        -------
        list
            threshold list

        """
        if len(masses) != 3:
            raise ValueError("There have to be 3 quark masses")
        if thresholds_ratios is None:
            thresholds_ratios = [1.0, 1.0, 1.0]
        if len(thresholds_ratios) != 3:
            raise ValueError("There have to be 3 quark threshold ratios")
        if max_nf is None:
            max_nf = 6

        thresholds = []
        for m, k in zip(masses, thresholds_ratios):
            thresholds.append(m * k)
        # cut array = simply reduce some thresholds
        thresholds = thresholds[: max_nf - 3]
        return thresholds

    def path(
        self,
        q2_to: float,
        nf_to: Optional[int] = None,
        q2_from: Optional[float] = None,
        nf_from: Optional[int] = None,
    ):
        """Get path from ``q2_from`` to ``q2_to``.

        Parameters
        ----------
        q2_to:
            target value of q2
        q2_from:
            starting value of q2

        Returns
        -------
        list(PathSegment)
            List of :class:`PathSegment` to go through in order to get from ``q2_from``
            to ``q2_to``

        """
        # fallback to init config
        if q2_from is None:
            q2_from = self.q2_ref
        if nf_from is None:
            nf_from = self.nf_ref
        # determine reference thresholds
        if nf_from is None:
            nf_from = 2 + np.digitize(q2_from, self.area_walls)
        if nf_to is None:
            nf_to = 2 + np.digitize(q2_to, self.area_walls)
        # determine direction and python slice modifier
        if nf_to < nf_from:
            rc = -1
            shift = -3
        else:
            rc = 1
            shift = -2
        # join all necessary points in one list
        boundaries = (
            [q2_from]
            + self.area_walls[nf_from + shift : int(nf_to) + shift : rc]
            + [q2_to]
        )
        segs = [
            PathSegment(boundaries[i], q2, nf_from + i * rc)
            for i, q2 in enumerate(boundaries[1:])
        ]
        return segs

    def nf(self, q2):
        """Find the number of flavors active at the given scale.

        Parameters
        ----------
        q2 : float
            reference scale

        Returns
        -------
        int
            number of active flavors

        """
        ref_idx = np.digitize(q2, self.area_walls)
        return 2 + ref_idx


def is_downward_path(path: List[PathSegment]) -> bool:
    """Determine if a path is downward.

    Criterias are:

    - in the number of active flavors when the path list contains more than one
      :class:`PathSegment`, note this can be different from each
      :attr:`PathSegment.is_downward_q2`
    - in :math:`Q^2` when just one single :class:`PathSegment` is given

    Parameters
    ----------
    path :
        path

    Returns
    -------
    bool
        True for a downward path

    """
    if len(path) == 1:
        return path[0].is_downward_q2
    return path[1].nf < path[0].nf


def flavor_shift(is_downward: bool) -> int:
    """Determine the shift to number of light flavors.

    Parameters
    ----------
    is_downward : bool
        True for a downward path

    Returns
    -------
    int
        shift to number of light flavors which can be 3 or 4

    """
    return 4 if is_downward else 3



"""Contains the :class:`PhysicalOperator` class."""

import numpy as np

from .. import basis_rotation as br
from .. import member
from . import flavors


class PhysicalOperator(member.OperatorBase):
    """
    Join several fixed flavor scheme operators together.

    - provides the connection between the 7-dimensional anomalous dimension
      basis and the 15-dimensional evolution basis
    - provides the connection between the 15-dimensional evolution basis
      and the 169-dimensional flavor basis


    Parameters
    ----------
        op_members : dict
            list of all members
        q2_final : float
            final scale
    """

    @classmethod
    def ad_to_evol_map(cls, op_members, nf, q2_final, intrinsic_range, qed=False):
        """
        Obtain map between the 3-dimensional anomalous dimension basis and the 4-dimensional evolution basis.

        .. todo:: in VFNS sometimes IC is irrelevant if nf>=4

        Parameters
        ----------
            op_members : dict
                operator members in anomalous dimension basis
            nf : int
                number of active light flavors
            intrinsic_range : sequence
                intrinsic heavy flavors
            qed : bool
                activate qed

        Returns
        -------
            m : dict
                map
        """
        # constant elements
        m = {
            "S.S": op_members[(100, 100)],
            "S.g": op_members[(100, 21)],
            "g.g": op_members[(21, 21)],
            "g.S": op_members[(21, 100)],
        }
        if not qed:
            m.update({"V.V": op_members[(br.non_singlet_pids_map["nsV"], 0)]})
            # add elements which are already active
            for f in range(2, nf + 1):
                n = f**2 - 1
                m[f"V{n}.V{n}"] = op_members[(br.non_singlet_pids_map["ns-"], 0)]
                m[f"T{n}.T{n}"] = op_members[(br.non_singlet_pids_map["ns+"], 0)]
        else:
            m.update(
                {
                    "g.ph": op_members[(21, 22)],
                    "g.Sdelta": op_members[(21, 101)],
                    "ph.g": op_members[(22, 21)],
                    "ph.ph": op_members[(22, 22)],
                    "ph.S": op_members[(22, 100)],
                    "ph.Sdelta": op_members[(22, 101)],
                    "S.ph": op_members[(100, 22)],
                    "S.Sdelta": op_members[(100, 101)],
                    "Sdelta.g": op_members[(101, 21)],
                    "Sdelta.ph": op_members[(101, 22)],
                    "Sdelta.S": op_members[(101, 100)],
                    "Sdelta.Sdelta": op_members[(101, 101)],
                    "V.V": op_members[(10200, 10200)],
                    "V.Vdelta": op_members[(10200, 10204)],
                    "Vdelta.V": op_members[(10204, 10200)],
                    "Vdelta.Vdelta": op_members[(10204, 10204)],
                }
            )
            # add elements which are already active
            if nf >= 3:
                m["Td3.Td3"] = op_members[(br.non_singlet_pids_map["ns+d"], 0)]
                m["Vd3.Vd3"] = op_members[(br.non_singlet_pids_map["ns-d"], 0)]
            if nf >= 4:
                m["Tu3.Tu3"] = op_members[(br.non_singlet_pids_map["ns+u"], 0)]
                m["Vu3.Vu3"] = op_members[(br.non_singlet_pids_map["ns-u"], 0)]
            if nf >= 5:
                m["Td8.Td8"] = op_members[(br.non_singlet_pids_map["ns+d"], 0)]
                m["Vd8.Vd8"] = op_members[(br.non_singlet_pids_map["ns-d"], 0)]
            if nf >= 6:
                m["Tu8.Tu8"] = op_members[(br.non_singlet_pids_map["ns+u"], 0)]
                m["Vu8.Vu8"] = op_members[(br.non_singlet_pids_map["ns-u"], 0)]
        # deal with intrinsic heavy quark PDFs
        if intrinsic_range is not None:
            hqfl = "cbt"
            op_id = member.OpMember.id_like(op_members[(21, 21)])
            for intr_fl in intrinsic_range:
                if intr_fl <= nf:  # light quarks are not intrinsic
                    continue
                hq = hqfl[intr_fl - 4]  # find name
                # intrinsic means no evolution, i.e. they are evolving with the identity
                m[f"{hq}+.{hq}+"] = op_id.copy()
                m[f"{hq}-.{hq}-"] = op_id.copy()
        # map key to MemberName
        return cls.promote_names(m, q2_final)

    def to_flavor_basis_tensor(self, qed=False):
        """
        Convert the computations into an rank 4 tensor over flavor operator space and momentum fraction operator space.

        Parameters
        ----------
            qed : bool
                activate qed

        Returns
        -------
            tensor : numpy.ndarray
                EKO
        """
        nf_in, nf_out = flavors.get_range(self.op_members.keys(), qed)
        len_pids = len(br.flavor_basis_pids)
        len_xgrid = list(self.op_members.values())[0].value.shape[0]
        # dimension will be pids^2 * xgrid^2
        value_tensor = np.zeros((len_pids, len_xgrid, len_pids, len_xgrid))
        error_tensor = value_tensor.copy()
        for name, op in self.op_members.items():
            if not qed:
                in_pids = flavors.pids_from_intrinsic_evol(name.input, nf_in, False)
                out_pids = flavors.pids_from_intrinsic_evol(name.target, nf_out, True)
            else:
                in_pids = flavors.pids_from_intrinsic_unified_evol(
                    name.input, nf_in, False
                )
                out_pids = flavors.pids_from_intrinsic_unified_evol(
                    name.target, nf_out, True
                )
            for out_idx, out_weight in enumerate(out_pids):
                for in_idx, in_weight in enumerate(in_pids):
                    # keep the outer index to the left as we're multiplying from the right
                    value_tensor[
                        out_idx,  # output pid (position)
                        :,  # output momentum fraction
                        in_idx,  # input pid (position)
                        :,  # input momentum fraction
                    ] += out_weight * (op.value * in_weight)
                    error_tensor[
                        out_idx,  # output pid (position)
                        :,  # output momentum fraction
                        in_idx,  # input pid (position)
                        :,  # input momentum fraction
                    ] += out_weight * (op.error * in_weight)
        return value_tensor, error_tensor



"""Manipulate output generate by EKO."""
import logging
import warnings
from typing import Optional

import numpy as np

from .. import basis_rotation as br
from .. import interpolation
from .struct import EKO

logger = logging.getLogger(__name__)

TARGETGRID_ROTATION = "ij,ajbk->aibk"
INPUTGRID_ROTATION = "ajbk,kl->ajbl"
SIMGRID_ROTATION = "ij,ajbk,kl->aibl"
"""Simultaneous grid rotation contraction indices."""


def xgrid_reshape(
    eko: EKO,
    targetgrid: Optional[interpolation.XGrid] = None,
    inputgrid: Optional[interpolation.XGrid] = None,
):
    """Reinterpolate operators on output and/or input grids.

    The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    targetgrid :
        xgrid for the target (output PDF)
    inputgrid :
        xgrid for the input (input PDF)

    """
    eko.assert_permissions(write=True)

    # calling with no arguments is an error
    if targetgrid is None and inputgrid is None:
        raise ValueError("Nor inputgrid nor targetgrid was given")
    # now check to the current status
    if (
        targetgrid is not None
        and len(targetgrid) == len(eko.rotations.targetgrid)
        and np.allclose(targetgrid.raw, eko.rotations.targetgrid.raw)
    ):
        targetgrid = None
        warnings.warn("The new targetgrid is close to the current targetgrid")
    if (
        inputgrid is not None
        and len(inputgrid) == len(eko.rotations.inputgrid)
        and np.allclose(inputgrid.raw, eko.rotations.inputgrid.raw)
    ):
        inputgrid = None
        warnings.warn("The new inputgrid is close to the current inputgrid")
    # after the checks: if there is still nothing to do, skip
    if targetgrid is None and inputgrid is None:
        logger.debug("Nothing done.")
        return

    # construct matrices
    if targetgrid is not None:
        b = interpolation.InterpolatorDispatcher(
            eko.rotations.targetgrid,
            eko.operator_card.configs.interpolation_polynomial_degree,
            False,
        )
        target_rot = b.get_interpolation(targetgrid.raw)
        eko.rotations.targetgrid = targetgrid
    if inputgrid is not None:
        b = interpolation.InterpolatorDispatcher(
            inputgrid,
            eko.operator_card.configs.interpolation_polynomial_degree,
            False,
        )
        input_rot = b.get_interpolation(eko.rotations.inputgrid.raw)
        eko.rotations.inputgrid = inputgrid

    # build new grid
    for q2, elem in eko.items():
        ops = elem.operator
        errs = elem.error
        if targetgrid is not None and inputgrid is None:
            ops = np.einsum(TARGETGRID_ROTATION, target_rot, ops, optimize="optimal")
            errs = (
                np.einsum(TARGETGRID_ROTATION, target_rot, errs, optimize="optimal")
                if errs is not None
                else None
            )
        elif inputgrid is not None and targetgrid is None:
            ops = np.einsum(INPUTGRID_ROTATION, ops, input_rot, optimize="optimal")
            errs = (
                np.einsum(INPUTGRID_ROTATION, errs, input_rot, optimize="optimal")
                if errs is not None
                else None
            )
        else:
            ops = np.einsum(
                SIMGRID_ROTATION, target_rot, ops, input_rot, optimize="optimal"
            )
            errs = (
                np.einsum(
                    SIMGRID_ROTATION, target_rot, errs, input_rot, optimize="optimal"
                )
                if errs is not None
                else None
            )
        elem.operator = ops
        elem.error = errs

        eko[q2] = elem

    eko.update()


TARGETPIDS_ROTATION = "ca,ajbk->cjbk"
INPUTPIDS_ROTATION = "ajbk,bd->ajdk"
SIMPIDS_ROTATION = "ca,ajbk,bd->cjdk"
"""Simultaneous grid rotation contraction indices."""


def flavor_reshape(
    eko: EKO,
    targetpids: Optional[np.ndarray] = None,
    inputpids: Optional[np.ndarray] = None,
    update: bool = True,
):
    """Change the operators to have in the output targetpids and/or in the input inputpids.

    The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    targetpids :
        target rotation specified in the flavor basis
    inputpids :
        input rotation specified in the flavor basis
    update :
        update :class:`~eko.io.struct.EKO` metadata after writing

    """
    eko.assert_permissions(write=True)

    # calling with no arguments is an error
    if targetpids is None and inputpids is None:
        raise ValueError("Nor inputpids nor targetpids was given")
    # now check to the current status
    if targetpids is not None and np.allclose(
        targetpids, np.eye(len(eko.rotations.targetpids))
    ):
        targetpids = None
        warnings.warn("The new targetpids is close to current basis")
    if inputpids is not None and np.allclose(
        inputpids, np.eye(len(eko.rotations.inputpids))
    ):
        inputpids = None
        warnings.warn("The new inputpids is close to current basis")
    # after the checks: if there is still nothing to do, skip
    if targetpids is None and inputpids is None:
        logger.debug("Nothing done.")
        return

    # flip input around
    if inputpids is not None:
        inv_inputpids = np.linalg.inv(inputpids)

    # build new grid
    for q2, elem in eko.items():
        ops = elem.operator
        errs = elem.error
        if targetpids is not None and inputpids is None:
            ops = np.einsum(TARGETPIDS_ROTATION, targetpids, ops, optimize="optimal")
            errs = (
                np.einsum(TARGETPIDS_ROTATION, targetpids, errs, optimize="optimal")
                if errs is not None
                else None
            )
        elif inputpids is not None and targetpids is None:
            ops = np.einsum(INPUTPIDS_ROTATION, ops, inv_inputpids, optimize="optimal")
            errs = (
                np.einsum(INPUTPIDS_ROTATION, errs, inv_inputpids, optimize="optimal")
                if errs is not None
                else None
            )
        else:
            ops = np.einsum(
                SIMPIDS_ROTATION, targetpids, ops, inv_inputpids, optimize="optimal"
            )
            errs = (
                np.einsum(
                    SIMPIDS_ROTATION,
                    targetpids,
                    errs,
                    inv_inputpids,
                    optimize="optimal",
                )
                if errs is not None
                else None
            )
        elem.operator = ops
        elem.error = errs

        eko[q2] = elem

    # drop PIDs - keeping them int nevertheless
    # there is no meaningful way to set them in general, after rotation
    if inputpids is not None:
        eko.rotations.inputpids = np.array([0] * len(eko.rotations.inputpids))
    if targetpids is not None:
        eko.rotations.targetpids = np.array([0] * len(eko.rotations.targetpids))

    if update:
        eko.update()


def to_evol(eko: EKO, source: bool = True, target: bool = False):
    """Rotate the operator into evolution basis.

    This also assigns also the pids. The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor

    """
    # rotate
    inputpids = br.rotate_flavor_to_evolution if source else None
    targetpids = br.rotate_flavor_to_evolution if target else None
    # prevent metadata update, since flavor_reshape has not enough information
    # to determine inpupids and targetpids, and they will be updated after the
    # call
    flavor_reshape(eko, inputpids=inputpids, targetpids=targetpids, update=False)
    # assign pids
    if source:
        eko.rotations.inputpids = inputpids
    if target:
        eko.rotations.targetpids = targetpids

    eko.update()


def to_uni_evol(eko: EKO, source: bool = True, target: bool = False):
    """Rotate the operator into evolution basis.

    This also assigns also the pids. The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor

    """
    # rotate
    inputpids = br.rotate_flavor_to_unified_evolution if source else None
    targetpids = br.rotate_flavor_to_unified_evolution if target else None
    # prevent metadata update, since flavor_reshape has not enough information
    # to determine inpupids and targetpids, and they will be updated after the
    # call
    flavor_reshape(eko, inputpids=inputpids, targetpids=targetpids, update=False)
    # assign pids
    if source:
        eko.rotations.inputpids = inputpids
    if target:
        eko.rotations.targetpids = targetpids

    eko.update()



"""Define output representation structures."""
import base64
import contextlib
import copy
import io
import logging
import os
import pathlib
import shutil
import tarfile
import tempfile
from dataclasses import dataclass
from typing import BinaryIO, Dict, Optional

import lz4.frame
import numpy as np
import numpy.lib.npyio as npyio
import numpy.typing as npt
import yaml

from .. import interpolation
from .. import version as vmod
from . import exceptions, raw
from .dictlike import DictLike
from .runcards import OperatorCard, Rotations, TheoryCard

logger = logging.getLogger(__name__)

THEORYFILE = "theory.yaml"
OPERATORFILE = "operator.yaml"
METADATAFILE = "metadata.yaml"
RECIPESDIR = "recipes"
PARTSDIR = "parts"
OPERATORSDIR = "operators"

COMPRESSED_SUFFIX = ".lz4"


@dataclass
class Operator(DictLike):
    """Operator representation.

    To be used to hold the result of a computed 4-dim operator (from a given
    scale to another given one).

    Note
    ----
    IO works with streams in memory, in order to avoid intermediate write on
    disk (keep read from and write to tar file only).

    """

    operator: npt.NDArray
    """Content of the evolution operator."""
    error: Optional[npt.NDArray] = None
    """Errors on individual operator elements (mainly used for integration
    error, but it can host any kind of error).
    """

    def save(self, stream: BinaryIO, compress: bool = True) -> bool:
        """Save content of operator to bytes.

        Parameters
        ----------
        stream : BinaryIO
            a stream where to save the operator content (in order to be able to
            perform the operation both on disk and in memory)
        compress : bool
            flag to control optional compression (default: `True`)

        Returns
        -------
        bool
            whether the operator saved contained or not the error (this control
            even the format, ``npz`` with errors, ``npy`` otherwise)

        """
        aux = io.BytesIO()
        if self.error is None:
            np.save(aux, self.operator)
        else:
            np.savez(aux, operator=self.operator, error=self.error)
        aux.seek(0)

        # compress if requested
        if compress:
            content = lz4.frame.compress(aux.read())
        else:
            content = aux.read()

        # write compressed data
        stream.write(content)
        stream.seek(0)

        # return the type of array dumped (i.e. 'npy' or 'npz')
        return self.error is None

    @classmethod
    def load(cls, stream: BinaryIO, compressed: bool = True):
        """Load operator from bytes.

        Parameters
        ----------
        stream : BinaryIO
            a stream to load the operator from (to support the operation both on
            disk and in memory)
        compressed : bool
            declare whether the stream is or is not compressed (default: `True`)

        Returns
        -------
        Operator
            the loaded instance

        """
        if compressed:
            stream = io.BytesIO(lz4.frame.decompress(stream.read()))
        content = np.load(stream)

        if isinstance(content, np.ndarray):
            op = content
            err = None
        elif isinstance(content, npyio.NpzFile):
            op = content["operator"]
            err = content["error"]
        else:
            raise exceptions.OperatorLoadingError(
                "Not possible to load operator, content format not recognized"
            )

        return cls(operator=op, error=err)


@dataclass
class InternalPaths:
    """Paths inside an EKO folder.

    This structure exists to locate in a single place the internal structure of
    an EKO folder.

    The only value required is the root path, everything else is computed
    relative to this root.
    In case only the relative paths are required, just create this structure
    with :attr:`root` equal to emtpty string or ``"."``.

    """

    root: pathlib.Path
    "The root of the EKO folder (use placeholder if not relevant)"

    @property
    def metadata(self):
        """Metadata file."""
        return self.root / METADATAFILE

    @property
    def recipes(self):
        """Recipes folder."""
        return self.root / RECIPESDIR

    @property
    def parts(self):
        """Parts folder."""
        return self.root / PARTSDIR

    @property
    def operators(self):
        """Operators folder.

        This is the one containing the actual EKO components, after
        computation has been performed.

        """
        return self.root / OPERATORSDIR

    @property
    def theory_card(self):
        """Theory card dump."""
        return self.root / THEORYFILE

    @property
    def operator_card(self):
        """Operator card dump."""
        return self.root / OPERATORFILE

    def bootstrap(self, theory: dict, operator: dict, metadata: dict):
        """Create directory structure.

        Parameters
        ----------
        dirpath : os.PathLike
            path to create the directory into
        theory : dict
            theory card to be dumped
        operator : dict
            operator card to be dumped
        metadata : dict
            metadata of the operator

        """
        self.metadata.write_text(yaml.dump(metadata), encoding="utf-8")
        self.theory_card.write_text(yaml.dump(theory), encoding="utf-8")
        self.operator_card.write_text(yaml.dump(operator), encoding="utf-8")
        self.recipes.mkdir()
        self.parts.mkdir()
        self.operators.mkdir()

    @staticmethod
    def opname(mu2: float) -> str:
        r"""Operator file name from :math:`\mu^2` value."""
        decoded = np.float64(mu2).tobytes()
        return base64.urlsafe_b64encode(decoded).decode()

    def oppath(self, mu2: float) -> pathlib.Path:
        r"""Retrieve operator file path from :math:`\mu^2` value.

        This method looks for an existing path matching.

        Parameters
        ----------
        mu2: float
            :math:`\mu2` scale specified

        Returns
        -------
        pathlib.Path
            the path retrieved, guaranteed to exist

        Raises
        ------
        ValueError
            if the path is not found, or more than one are matching the
            specified value of :math:`\mu2`

        """
        name = self.opname(mu2)
        oppaths = list(
            filter(lambda path: path.name.startswith(name), self.operators.iterdir())
        )

        if len(oppaths) == 0:
            raise ValueError(f"mu2 value '{mu2}' not available.")
        elif len(oppaths) > 1:
            raise ValueError(f"Too many operators associated to '{mu2}':\n{oppaths}")

        return oppaths[0]

    def opcompressed(self, path: os.PathLike) -> bool:
        """Check if the operator at the path specified is compressed.

        Parameters
        ----------
        path: os.PathLike
            the path to the operator to check

        Returns
        -------
        bool
            whether it is compressed

        Raises
        ------
        OperatorLocationError
            if the path is not inside the operators folder

        """
        path = pathlib.Path(path)
        if self.operators not in path.parents:
            raise exceptions.OperatorLocationError(path)

        return path.suffix == COMPRESSED_SUFFIX

    def opmu2(self, path: os.PathLike) -> float:
        r"""Extract :math:`\mu2` value from operator path.

        Parameters
        ----------
        path: os.PathLike
            the path to the operator

        Returns
        -------
        bool
            the :math:`\mu2` value associated to the operator

        Raises
        ------
        OperatorLocationError
            if the path is not inside the operators folder

        """
        path = pathlib.Path(path)
        if self.operators not in path.parents:
            raise exceptions.OperatorLocationError(path)

        encoded = path.stem.split(".")[0]
        decoded = base64.urlsafe_b64decode(encoded)
        return float(np.frombuffer(decoded, dtype=np.float64)[0])

    def opnewpath(
        self, mu2: float, compress: bool = True, without_err: bool = True
    ) -> pathlib.Path:
        r"""Compute the path associated to :math:`\mu^2` value.

        Parameters
        ----------
        mu2: float
            :math:`\mu2` scale specified

        Returns
        -------
        pathlib.Path
            the path computed, it might already exists

        """
        suffix = ".npy" if without_err else ".npz"
        if compress:
            suffix += COMPRESSED_SUFFIX
        return self.operators / (self.opname(mu2) + suffix)

    @property
    def mu2grid(self) -> npt.NDArray[np.float64]:
        r"""Provide the array of :math:`\mu2` values of existing operators."""
        if self.root is None:
            raise RuntimeError()

        return np.array(
            sorted(
                map(
                    lambda p: self.opmu2(p),
                    InternalPaths(self.root).operators.iterdir(),
                )
            )
        )


@dataclass
class AccessConfigs:
    """Configurations specified during opening of an EKO."""

    path: pathlib.Path
    """The path to the permanent object."""
    readonly: bool
    "Read-only flag"
    open: bool
    "EKO status"

    def assert_open(self):
        """Assert operator is open.

        Raises
        ------
        exceptions.ClosedOperator
            if operator is closed

        """
        if not self.open:
            raise exceptions.ClosedOperator

    def assert_writeable(self, msg: Optional[str] = None):
        """Assert operator is writeable.

        Raises
        ------
        exceptions.ClosedOperator
            see :meth:`assert_open`
        exceptions.ReadOnlyOperator
            if operators has been declared read-only

        """
        if msg is None:
            msg = ""

        self.assert_open()
        if self.readonly:
            raise exceptions.ReadOnlyOperator(msg)

    @property
    def read(self):
        """Check reading permission.

        Reading access is always granted on open operator.

        """
        return self.open

    @property
    def write(self):
        """Check writing permission."""
        return self.open and not self.readonly


@dataclass
class Metadata(DictLike):
    """Manage metadata, and keep them synced on disk.

    It is possible to have a metadata view, in which the path is not actually
    connected (i.e. it is set to ``None``). In this case, no update will be
    possible, of course.

    Note
    ----
    Unfortunately, for nested structures it is not possible to detect a change
    in their attributes, so a call to :meth:`update` has to be performed
    manually.

    """

    mu20: float
    """Inital scale."""
    rotations: Rotations
    """Manipulation information, describing the current status of the EKO (e.g.
    `inputgrid` and `targetgrid`).
    """
    # tagging information
    _path: Optional[pathlib.Path] = None
    """Path to temporary dir."""
    version: str = vmod.__version__
    """Library version used to create the corresponding file."""
    data_version: int = vmod.__data_version__
    """Specs version, to which the file adheres."""

    @classmethod
    def load(cls, path: os.PathLike):
        """Load metadata from open folder.

        Parameters
        ----------
        path: os.PathLike
            the path to the open EKO folder

        Returns
        -------
        bool
            loaded metadata

        """
        path = pathlib.Path(path)
        content = cls.from_dict(
            yaml.safe_load(InternalPaths(path).metadata.read_text(encoding="utf-8"))
        )
        content._path = path
        return content

    def update(self):
        """Update the disk copy of metadata."""
        if self._path is None:
            logger.info("Impossible to set metadata, no file attached.")
        else:
            with open(InternalPaths(self._path).metadata, "w") as fd:
                yaml.safe_dump(self.raw, fd)

    @property
    def path(self):
        """Access temporary dir path.

        Raises
        ------
        RuntimeError
            if path has not been initialized before

        """
        if self._path is None:
            raise RuntimeError(
                "Access to EKO directory attempted, but not dir has been set."
            )
        return self._path

    @path.setter
    def path(self, value: pathlib.Path):
        """Set temporary dir path."""
        self._path = value

    @property
    def raw(self):
        """Override default :meth:`DictLike.raw` representation to exclude path."""
        return self.public_raw


@dataclass
class EKO:
    """Operator interface.

    This class offers an interface to an abstract operator, between memory and
    disk.

    An actual operator might be arbitrarily huge, and in particular size
    limitations in memory are far more strict than on disk.
    Since manually managing, for each application, the burden of off-loading
    part of the operator might be hard and occasionally not possible (without a
    clear picture of the internals), the library itself offers this facility.

    In particular, the data format on disk has a complete specification, and
    can hold a full operator independently of the loading procedure.
    In order to accomplish the former goal, the remaining task of partial
    loading is done by this class (for the Python library, other
    implementations are possible and encouraged).

    For this reason, a core component of an :class:`EKO` object is a path,
    referring to the location on disk of the corresponding operator.
    Any :class:`EKO` has an associated path:

        - for the computed object, it corresponds to the path where the actual
          result of the computation is already saved
        - for a new object, it is the path at which any result of final or
          intermediate computation is stored, as soon as it is produced

    The computation can be stopped at any time, without the loss of any of the
    intermediate results.

    """

    # operators cache, contains the Q2 grid information
    _operators: Dict[float, Optional[Operator]]

    # public containers
    # -----------------
    metadata: Metadata
    """Operator metadata."""
    access: AccessConfigs
    """Access related configurations."""

    # shortcut properties
    # -------------------
    @property
    def paths(self) -> InternalPaths:
        """Accessor for internal paths."""
        return InternalPaths(self.metadata.path)

    @property
    def rotations(self) -> Rotations:
        """Rotations information."""
        return self.metadata.rotations

    @property
    def xgrid(self) -> interpolation.XGrid:
        """Momentum fraction internal grid."""
        return self.rotations.xgrid

    @xgrid.setter
    def xgrid(self, value: interpolation.XGrid):
        """Set `xgrid` value."""
        self.rotations.xgrid = value
        self.update()

    @property
    def mu20(self) -> float:
        """Provide squared initial scale."""
        return self.metadata.mu20

    @property
    def mu2grid(self) -> npt.NDArray:
        """Provide the list of :math:`Q^2` as an array."""
        return np.array(list(self._operators))

    @property
    def theory_card(self):
        """Provide theory card, retrieving from the dump."""
        return TheoryCard.from_dict(
            yaml.safe_load(self.paths.theory_card.read_text(encoding="utf-8"))
        )

    @property
    def operator_card(self):
        """Provide operator card, retrieving from the dump."""
        return OperatorCard.from_dict(
            yaml.safe_load(self.paths.operator_card.read_text(encoding="utf-8"))
        )

    # persistency control
    # -------------------

    def update(self):
        """Write updates to structure for persistency."""
        self.access.assert_writeable()
        self.metadata.update()

    def assert_permissions(self, read=True, write=False):
        """Assert permissions on current operator."""
        if read:
            self.access.assert_open()
        if write:
            self.access.assert_writeable()

    @property
    def permissions(self):
        """Provide permissions information."""
        return dict(read=self.access.read, write=self.access.write)

    # operator management
    # -------------------
    def __getitem__(self, mu2: float) -> Operator:
        r"""Retrieve operator for given :math:`\mu^2`.

        If the operator is not already in memory, it will be automatically
        loaded.

        Parameters
        ----------
        mu2 : float
            :math:`\mu^2` value labeling the operator to be retrieved

        Returns
        -------
        Operator
            the retrieved operator

        """
        self.access.assert_open()

        if mu2 in self._operators:
            op = self._operators[mu2]
            if op is not None:
                return op

        oppath = self.paths.oppath(mu2)
        compressed = self.paths.opcompressed(oppath)

        with open(oppath, "rb") as fd:
            op = Operator.load(fd, compressed=compressed)

        self._operators[mu2] = op
        return op

    def __setitem__(self, mu2: float, op: Operator, compress: bool = True):
        """Set operator for given :math:`Q^2`.

        The operator is automatically dumped on disk.

        Parameters
        ----------
        q2 : float
            :math:`Q^2` value labeling the operator to be set
        op : Operator
            the retrieved operator
        compress : bool
            whether to save the operator compressed or not (default: `True`)

        """
        self.access.assert_writeable()

        if not isinstance(op, Operator):
            raise ValueError("Only an Operator can be added to an EKO")

        without_err = op.error is None
        oppath = self.paths.opnewpath(mu2, compress=compress, without_err=without_err)

        with open(oppath, "wb") as fd:
            without_err2 = op.save(fd, compress)
        assert without_err == without_err2

        self._operators[mu2] = op

    def __delitem__(self, q2: float):
        """Drop operator from memory.

        This method only drops the operator from memory, and it's not expected
        to do anything else.

        Autosave is done on set, and explicit saves are performed by the
        computation functions.

        If a further explicit save is required, repeat explicit assignment::

            eko[q2] = eko[q2]

        This is only useful if the operator has been mutated in place, that in
        general should be avoided, since the operator should only be the result
        of a full computation or a library manipulation.


        Parameters
        ----------
        q2 : float
            the value of :math:`Q^2` for which the corresponding operator
            should be dropped

        """
        self._operators[q2] = None

    @contextlib.contextmanager
    def operator(self, q2: float):
        """Retrieve an operator and discard it afterwards.

        To be used as a contextmanager: the operator is automatically loaded as
        usual, but on the closing of the context manager it is dropped from
        memory.

        Parameters
        ----------
        q2 : float
            :math:`Q^2` value labeling the operator to be retrieved

        """
        try:
            yield self[q2]
        finally:
            del self[q2]

    def __iter__(self):
        """Iterate over keys (i.e. Q2 values).

        Yields
        ------
        float
            q2 values

        """
        yield from self._operators

    def items(self):
        """Iterate operators, with minimal load.

        Pay attention, this iterator:

        - is not a read-only operation from the point of view of the in-memory
          object (since the final result after iteration is no operator loaded)
        - but it is a read-only operation from the point of view of the
          permanent object on-disk

        Yields
        ------
        tuple
            couples of ``(q2, operator)``, loaded immediately before, unloaded
            immediately after

        """
        for q2 in self.mu2grid:
            yield q2, self[q2]
            del self[q2]

    def __contains__(self, q2: float) -> bool:
        """Check whether :math:`Q^2` operators are present.

        'Present' means, in this case, they are conceptually part of the
        :class:`EKO`. But it is telling nothing about being loaded in memory or
        not.

        Returns
        -------
        bool
            the result of checked condition

        """
        return q2 in self._operators

    def approx(
        self, q2: float, rtol: float = 1e-6, atol: float = 1e-10
    ) -> Optional[float]:
        """Look for close enough :math:`Q^2` value in the :class:`EKO`.

        Parameters
        ----------
        q2 : float
            value of :math:`Q2` in which neighbourhood to look
        rtol : float
            relative tolerance
        atol : float
            absolute tolerance

        Returns
        -------
        float or None
            retrieved value of :math:`Q^2`, if a single one is found

        Raises
        ------
        ValueError
            if multiple values are find in the neighbourhood

        """
        q2s = self.mu2grid
        close = q2s[np.isclose(q2, q2s, rtol=rtol, atol=atol)]

        if close.size == 1:
            return close[0]
        if close.size == 0:
            return None
        raise ValueError(f"Multiple values of Q2 have been found close to {q2}")

    def unload(self):
        """Fully unload the operators in memory."""
        for q2 in self:
            del self[q2]

    # operator management
    # -------------------

    def deepcopy(self, path: os.PathLike):
        """Create a deep copy of current instance.

        The managed on-disk object is copied as well, to the new ``path``
        location.
        If you don't want to copy the disk, consider using directly::

            copy.deepcopy(myeko)

        It will perform the exact same operation, without propagating it to the
        disk counterpart.

        Parameters
        ----------
        path :
            path to the permanent location of the new object (not the temporary
            directory)

        Returns
        -------
        EKO
            the copy created

        """
        # deepcopy is tuned to copy an open operator into a closed one, since
        # operators are always created closed (to avoid delegating the user to
        # close an object he never opened), then let's assert this assumption
        self.access.assert_open()

        # return the object fully on disk, in order to avoid expensive copies
        # also in memory (they would be copied on-disk in any case)
        self.unload()

        new = copy.deepcopy(self)
        new.access.path = pathlib.Path(path)
        new.access.readonly = False
        new.access.open = True

        tmpdir = pathlib.Path(tempfile.mkdtemp())
        new.metadata.path = tmpdir
        # copy old dir to new dir
        tmpdir.rmdir()
        shutil.copytree(self.paths.root, new.paths.root)
        new.close()

    @staticmethod
    def load(tarpath: os.PathLike, dest: os.PathLike):
        """Load the content of archive in a target directory.

        Parameters
        ----------
        tarpath: os.PathLike
            the archive to extract
        tmppath: os.PathLike
            the destination directory

        """
        try:
            with tarfile.open(tarpath) as tar:
                raw.safe_extractall(tar, dest)
        except tarfile.ReadError:
            raise exceptions.OutputNotTar(f"Not a valid tar archive: '{tarpath}'")

    @classmethod
    def open(cls, path: os.PathLike, mode="r"):
        """Open EKO object in the specified mode."""
        path = pathlib.Path(path)
        access = AccessConfigs(path, readonly=False, open=True)
        load = False
        if mode == "r":
            load = True
            access.readonly = True
        elif mode in "w":
            pass
        elif mode in "a":
            load = True
        else:
            raise ValueError

        tmpdir = pathlib.Path(tempfile.mkdtemp())
        if load:
            cls.load(path, tmpdir)
            metadata = Metadata.load(tmpdir)
            opened = cls(
                _operators={mu2: None for mu2 in InternalPaths(metadata.path).mu2grid},
                metadata=metadata,
                access=access,
            )
        else:
            opened = Builder(path=tmpdir, access=access)

        return opened

    @classmethod
    def read(cls, path: os.PathLike):
        """Read the content of an EKO.

        Type-safe alias for::

            EKO.open(... , "r")

        """
        eko = cls.open(path, "r")
        assert isinstance(eko, EKO)
        return eko

    @classmethod
    def create(cls, path: os.PathLike):
        """Create a new EKO.

        Type-safe alias for::

            EKO.open(... , "w")

        """
        builder = cls.open(path, "w")
        assert isinstance(builder, Builder)
        return builder

    @classmethod
    def edit(cls, path: os.PathLike):
        """Read from and write on existing EKO.

        Type-safe alias for::

            EKO.open(... , "a")

        """
        eko = cls.open(path, "a")
        assert isinstance(eko, EKO)
        return eko

    def __enter__(self):
        """Allow EKO to be used in :obj:`with` statements."""
        return self

    def dump(self, archive: Optional[os.PathLike] = None):
        """Dump the current content to archive.

        Parameters
        ----------
        archive: os.PathLike or None
            path to archive, in general you should keep the default, that will
            make use of the registered path (default: ``None``)

        Raises
        ------
        ValueError
            when trying to dump on default archive in read-only mode

        """
        if archive is None:
            self.access.assert_writeable(
                "Not possible to dump on default archive in read-only mode"
            )
            archive = self.access.path

        with tarfile.open(archive, "w") as tar:
            tar.add(self.metadata.path, arcname=".")

    def close(self):
        """Close the current object, cleaning up.

        If not in read-only mode, dump to permanent storage.
        Remove the temporary directory used.

        """
        if not self.access.readonly:
            # clean given path, to overwrite it - default 'w'rite behavior
            self.access.path.unlink(missing_ok=True)
            self.dump()

        self.access.open = False
        shutil.rmtree(self.metadata.path)

    def __exit__(self, exc_type: type, _exc_value, _traceback):
        """Ensure EKO to be closed properly."""
        if exc_type is not None:
            return

        self.close()

    @property
    def raw(self) -> dict:
        """Provide raw representation of the full content.

        Returns
        -------
        dict
            nested dictionary, storing all the values in the structure, but the
            operators themselves

        """
        return dict(mu2grid=self.mu2grid.tolist(), metadata=self.metadata.raw)


@dataclass
class Builder:
    """Build EKO instances."""

    path: pathlib.Path
    """Path on disk to ."""
    access: AccessConfigs
    """Access related configurations."""

    # optional arguments, required at build time
    theory: Optional[TheoryCard] = None
    operator: Optional[OperatorCard] = None

    eko: Optional[EKO] = None

    def __post_init__(self):
        """Validate paths."""
        if self.access.path.suffix != ".tar":
            raise exceptions.OutputNotTar(self.access.path)
        if self.access.path.exists():
            raise exceptions.OutputExistsError(self.access.path)

    def load_cards(self, theory: TheoryCard, operator: OperatorCard):
        """Load both theory and operator card."""
        self.theory = theory
        self.operator = operator

        return self

    def build(self) -> EKO:
        """Build EKO instance.

        Returns
        -------
        EKO
            the constructed instance

        Raises
        ------
        RuntimeError
            if not enough information is available (at least one card missing)

        """
        missing = []
        for card in ["theory", "operator"]:
            if getattr(self, card) is None:
                missing.append(card)

        if len(missing) > 0:
            raise RuntimeError(
                f"Can not build an EKO, since following cards are missing: {missing}"
            )

        # tell the static analyzer as well
        assert self.theory is not None
        assert self.operator is not None

        self.access.open = True
        metadata = Metadata(
            _path=self.path,
            mu20=self.operator.mu20,
            rotations=Rotations(xgrid=self.operator.xgrid),
        )
        InternalPaths(self.path).bootstrap(
            theory=self.theory.raw,
            operator=self.operator.raw,
            metadata=metadata.raw,
        )

        self.eko = EKO(
            _operators={mu2: None for mu2 in self.operator.mu2grid},
            metadata=metadata,
            access=self.access,
        )

        return self.eko

    def __enter__(self):
        """Allow Builder to be used in :obj:`with` statements."""
        return self

    def __exit__(self, exc_type: type, _exc_value, _traceback):
        """Ensure EKO to be closed properly."""
        if exc_type is not None:
            return

        # assign to variable to help type checker, otherwise self.eko might be
        # a property, and its type can change at every evaluation
        eko = self.eko
        if eko is not None:
            eko.close()



"""Recipes containing instructions for atomic computations."""
from abc import ABC
from dataclasses import dataclass

from .. import EKO
from .. import scale_variations as sv
from ..io import runcards
from ..io.dictlike import DictLike
from ..io.types import SquaredScale
from ..thresholds import ThresholdsAtlas


@dataclass
class Recipe(DictLike, ABC):
    """Base recipe structure."""

    name: str


@dataclass
class EvolutionRecipe(Recipe):
    """Recipe compute evolution with a fixed number of light flavors."""

    final: bool
    mu20: SquaredScale
    mu2: SquaredScale


@dataclass
class MatchingRecipe(Recipe):
    """Recipe to compute the matching between two different flavor number schemes."""

    mu2: SquaredScale


def create(eko: EKO):
    """Create all associated recipes."""
    _ = eko.theory_card.matching

    masses = runcards.masses(
        eko.theory_card, eko.operator_card.configs.evolution_method
    )

    tc = ThresholdsAtlas(
        masses=masses,
        q2_ref=eko.operator_card.mu20,
        nf_ref=eko.theory_card.num_flavs_init,
        thresholds_ratios=None,
        max_nf=eko.theory_card.num_flavs_max_pdf,
    )

    for mu2 in eko.mu2grid:
        expanded = eko.operator_card.configs.scvar_method is sv.Modes.expanded
        mu2f = mu2 * eko.theory_card.xif**2 if expanded else mu2
